/***************************************************************************
                          gamefield.cpp  -  description
                             -------------------
    begin                : Tue Apr 4 2000
    copyright            : (C) 2000 by Frank von Daak
    email                : frank.fun@kpage.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <math.h>
#include <stdlib.h>
#include <time.h>

#include <qdir.h>
#include <qimage.h>
#include <qrect.h>
#include <qpainter.h>
#include <qtimer.h>
#include <qcanvas.h>

#include <kapp.h>
#include <kaudioplayer.h>
#include <kstddirs.h>
#include <klocale.h>
#include <kdebug.h>

#include "MyQCanvasView.h"
#include "sprites.h"

#include "gamefield.h"


#define WIDTH 800
#define HEIGHT 550

#define PREFIX kapp->dirs()->findResourceDir("appdata", "sprites/")

#define BULLET_IMAGE	"sprites/bullets/bullet-%1.ppm"
#define BULLET_MASK	"sprites/bullets/bullet-%1.pbm"
#define BULLET_FRAMES	24

#define GNATKILL_IMAGE	"sprites/mueke/mueked-%1.ppm"
#define GNATKILL_MASK	"sprites/mueke/mueked-%1.pbm"
#define GNATKILL_FRAMES	10

#define TUXKILL_IMAGE	"sprites/dtux/dtux0000.ppm"
#define TUXKILL_MASK	"sprites/dtux/dtux0000.pbm"
#define TUXKILL_FRAMES	1

#define SHOT_WAV "sounds/shot.wav"
#define EMPTY_WAV "sounds/leer.wav"
#define LOAD_WAV "sounds/laden.wav"

#define MAG_IMAGE "sprites/maggot/magl-%1.ppm"
#define MAG_MASK "sprites/maggot/magl-%1.pbm"
#define MAG_FRAMES 29

#define TUX_IMAGE "sprites/tux/tuxl-%1.ppm"
#define TUX_MASK  "sprites/tux/tuxl-%1.pbm"
#define TUX_FRAMES 12

#define REFRESH_DELAY  1

#define TEXT_SPEED     6

GameField::GameField(QWidget *parent, const char *name ) : QWidget(parent,name)
{
 resize(WIDTH, HEIGHT);

 bullets.setAutoDelete(true);
 gnats.setAutoDelete(true);
 dgnats.setAutoDelete(true);
 mags.setAutoDelete(true);
 tuxs.setAutoDelete(true);
 dtuxs.setAutoDelete(true);
	
 fc = 0;
 ft = 0;
 shots = 0;
 shootWait = 100;
 started = false;
 rnd = 0;

 time_t t = time(&t);
 srand((unsigned int) t);

 bulletImg = new QCanvasPixmapArray(PREFIX+BULLET_IMAGE, BULLET_FRAMES);
 for (int i = 0; i < BULLET_FRAMES; i++) {
	bulletImg->image(i)->setMask(bulletImg->image(i)->createHeuristicMask());
 }
	
 bulletReady = true;
 bulletTimer = new QTimer(this);
 connect(bulletTimer, SIGNAL(timeout()), this, SLOT(bulletTimeout()));
}

GameField::~GameField()
{
//clear the lists first
 bullets.clear();
 gnats.clear();
 dgnats.clear();
 mags.clear();
 tuxs.clear();
 dtuxs.clear();

// and now the rest
 delete field;
 delete view;
}

void GameField::addBullets(int num)
{
 for (int i = 0; i < num; i++) {
	double x = this->width() - (num + 1) * 25 + i * 25;
	Bullet* bullet = new Bullet(bulletImg, field);
	bullet->setVisible(true);
	bullet->setVelocity(-25, 0);
	bullet->move(x, bullet->y());
	bullets.append(bullet);
 }
}

void GameField::removeBullets(int num)//num???
{
 QCanvasSprite* bullet;
 bullet = bullets.first();
 bullet->hide();
 bullets.remove(bullet);
}

void GameField::bulletTimeout()
{
 bulletReady = false;
 bulletTimer->stop();

 QCanvasSprite* bullet;
 bullet = bullets.first();
 bullet->advance(1);
 bullet->setFrame( (bullet->frame()+1) % BULLET_FRAMES);

 if (bullet->x() <= 1) {
	bullet->hide();
	bullets.removeRef(bullet);
	bulletReady = true;
 } else 
	bulletTimer->start(1);
}

void GameField::setLevel(int lev)
{
 gnatSpeed = (lev/2)+2;
 switch ( random()%3 ){
	case 0:
		IMG_BACKGROUND = "sprites/background/bg.ppm";
		break;
	case 1:
		IMG_BACKGROUND = "sprites/background/bg2.ppm";
		break;
	case 2:
		IMG_BACKGROUND = "sprites/background/bg3.ppm";
		break;
	default:
		IMG_BACKGROUND = "sprites/background/bg.ppm";
		break;
 }

 L_MUEKE = "sprites/mueke/muekel-%1.ppm";
 L_MMASK = "sprites/mueke/muekel-%1.pbm";
 L_MFRAMES = 32;
	
 R_MUEKE = "sprites/mueke/mueker-%1.ppm";
 R_MMASK = "sprites/mueke/mueker-%1.pbm";
 R_MFRAMES = 32;

 L_BUCK  = "sprites/buck/buckl-%1.ppm";
 L_BMASK  = "sprites/buck/buckl-%1.pbm";
 L_BFRAMES = 24;

 R_BUCK = "sprites/buck/buckr-%1.ppm";
 R_BMASK = "sprites/buck/buckr-%1.pbm";
 R_BFRAMES = 24;

 field = new QCanvas(WIDTH, HEIGHT);
 field->setBackgroundPixmap(QPixmap(locate("appdata", IMG_BACKGROUND)));

 view = new MyQCanvasView(field, this);
 view->setHScrollBarMode(QScrollView::AlwaysOff);
 view->setVScrollBarMode(QScrollView::AlwaysOff);
 view->resize(WIDTH, HEIGHT);//why?->should be in constructor 

 textSprite = new QCanvasText("", field);
 pauseTextSprite = new QCanvasText("", field);

 readSprites();
 connect(view, SIGNAL(fire(QPoint)), this, SLOT(shoot(QPoint)));
 connect(view, SIGNAL(load()), this, SLOT(reloadGun()));
}

void GameField::showPause(bool paused)
{
 if ( paused ) {
	pauseTextSprite->setText(i18n("\tGame paused !\nPress P to continue..."));
	pauseTextSprite->setFont(QFont("helvetica", 24, QFont::Bold));
	pauseTextSprite->setColor( yellow );
	pauseTextSprite->move( 350,270 );
	pauseTextSprite->show();
	field->update();
 } else {
	pauseTextSprite->hide();
	field->update();
 }
}

void GameField::readSprites()
{
//    cloudImg = new QCanvasPixmapArray(PREFIX + CLOUD1_IMAGE), CLOUD1_FRAMES );

// you could use:
//	tuxImg->image(i)->setMask(tuxImg->image(i)->createHeuristicMask());
// but I don't use it here because of performance. (long loading time)

 QString s;
 leftMImg = new QCanvasPixmapArray(PREFIX + L_MUEKE, L_MFRAMES );
 for (int i= 0; i < L_MFRAMES; i++) {
	s.sprintf("%04d", i);
    	leftMImg->image(i)->setMask(PREFIX + QString(L_MMASK).arg(s));
 }

 rightMImg = new QCanvasPixmapArray(PREFIX + R_MUEKE, R_MFRAMES );
 for (int i= 0; i < R_MFRAMES; i++) {
	s.sprintf("%04d", i);
    	rightMImg->image(i)->setMask(PREFIX + QString(R_MMASK).arg(s));
 }

 leftBImg = new QCanvasPixmapArray(PREFIX + L_BUCK, L_BFRAMES );
 for (int i= 0; i < L_BFRAMES; i++) {
	s.sprintf("%04d", i);
    	leftBImg->image(i)->setMask(PREFIX + QString(L_BMASK).arg(s));
 }

 rightBImg = new QCanvasPixmapArray(PREFIX + R_BUCK, R_BFRAMES );
 for (int i= 0; i < R_BFRAMES; i++) {
	s.sprintf("%04d", i);
    	rightBImg->image(i)->setMask(PREFIX + QString(R_BMASK).arg(s));
 }

 GnatKillImg = new QCanvasPixmapArray(PREFIX + GNATKILL_IMAGE,
		GNATKILL_FRAMES );
 for (int i= 0; i < GNATKILL_FRAMES; i++) {
	s.sprintf("%04d", i);
    	GnatKillImg->image(i)->setMask(PREFIX + QString(GNATKILL_MASK).arg(s));
 }

 TuxKillImg = new QCanvasPixmapArray(PREFIX + TUXKILL_IMAGE,
		TUXKILL_FRAMES );
 TuxKillImg->image(0)->setMask(TuxKillImg->image(0)->createHeuristicMask());
 TuxKillImg->image(0)->setMask(PREFIX + TUXKILL_MASK);

 magImg = new QCanvasPixmapArray(PREFIX + MAG_IMAGE, MAG_FRAMES );
 for (int i= 0; i < MAG_FRAMES; i++) {
	s.sprintf("%04d", i);
    	magImg->image(i)->setMask(PREFIX + QString(MAG_MASK).arg(s));
 }

 tuxImg = new QCanvasPixmapArray(PREFIX + TUX_IMAGE, TUX_FRAMES );
 for (int i= 0; i < TUX_FRAMES; i++) {
	s.sprintf("%04d", i);
    	tuxImg->image(i)->setMask(PREFIX + QString(TUX_MASK).arg(s));
 }

}

void GameField::showText( QString text, const QColor &color )
{
 textSprite->setText( text );
 textSprite->setFont(QFont("helvetica", 32, QFont::Bold));
 textSprite->setColor( color );

 textDy = TEXT_SPEED;

 textSprite->move( (width() - textSprite->boundingRect().width()) / 2, -5 );

 textSprite->show();
}

void GameField::newGame()
{
 reset();
 startTimer( 40 );
}

void GameField::reset()
{
 emit gameReset(shots);
 shots = 0;
}

void GameField::timerEvent( QTimerEvent * )
{
// time critical method!
// try to do everything fast here
 killTimers();

 shootWait++;

// move insects forward
 QListIterator<QCanvasSprite> it(mags);
 while (it.current()) {
	QCanvasSprite *mag = it.current();
	mag->advance(1);
	if ( mag->x() > 790 ) {
		mags.removeRef(mag);
	} else 
		mag->setFrame( ( mag->frame()+1 ) % MAG_FRAMES );
	++it;
 }

 it = tuxs;
 while (it.current()) {
	QCanvasSprite *tux = it.current();
	tux->advance( 1 );
	tux->setFrame( ( tux->frame()+1 ) % TUX_FRAMES );
	if ( tux->x() > 790 ) {
		tuxs.removeRef(tux);
	}

	++it;
 }

 it = gnats;
 while (it.current()) {
	QCanvasSprite *gnat = it.current();
	gnat->advance(1);
	gnat->setFrame( (gnat->frame() + 1) % 24 );
	if ( !started )
		gnat->hide();
	else
		gnat->show();

		//was QwRealMobileSprite::Wrap - is this possible with QCanvas?
	if (gnat->x() < 0-gnat->boundingRect().width()) {
		gnat->move(field->width(), gnat->y());
	} else if (gnat->x() > field->width()) {
		gnat->move(0 - gnat->boundingRect().width(), gnat->y());
	}

	if (gnat->y() < 0 - gnat->boundingRect().height()) {
		gnat->move(gnat->x(), field->height());
	} else if (gnat->y() > height()) {
		gnat->move(gnat->x(), 0 - gnat->boundingRect().height());
	}
	++it;
 }

 // Let them splash :-)
 it = dgnats;
 while(it.current()) {
	QCanvasSprite *dgnat = it.current();
	switch ( dgnat->rtti() ) {
		case GNAT_DEAD:
			fc++;
			dgnat->setFrame( ( dgnat->frame()+1 ) %
					GNATKILL_FRAMES );
			if ( fc >= dgnat->frameCount() )
				removeDGnat();
			break;
	}
	++it;
 }

 // Let them splash :-)
 it = dtuxs;
 while (it.current()) {
	QCanvasSprite *dtux = it.current();
	ft++;
	dtux->setFrame( ( dtux->frame() ) % 1 );
	if ( ft >= 32 )
		removeDTux();
	++it;
 }

 if ( !started && textSprite->visible() ) {
	if ( textDy < 0 && textSprite->boundingRect().y() <= -5 ) {
	    textSprite->hide();
	} else {
	    textSprite->moveBy( 0, textDy );
	}
	if ( textSprite->boundingRect().y() > height()/2 ) {
	    textDy = 0;
	}
 } else {
	textSprite->hide();
//	delete(textSprite);
 }

 field->update();

 startTimer( 40 );
}

void GameField::addGnats( int num )
{
 int i;
 double xmov, ymov;

 KGnat *gnat;
 for ( i = 0; i < num; i++ ) {
	xmov = (2.0 - (double)(random()%256)/64.0);
	ymov = (2.0 - (double)(random()%256)/64.0);

	if ( xmov > 0 ) {
		switch ( random() %2 ) {
			case 0:
				gnat = new KGnat( leftMImg, field );
				break;
			case 1:
				gnat = new KGnat( leftBImg, field );
				break;
			default:
				gnat = new KGnat( leftMImg, field );
		}
	} else {
		switch ( random() %2) {
			case 0:
				gnat = new KGnat( rightMImg, field );
				break;
			case 1:
				gnat = new KGnat( rightBImg, field );
				break;
			default:
				gnat = new KGnat( rightMImg, field );
		}
	}
		
//	gnat->setBoundsAction( QwRealMobileSprite::Wrap );//FIXME->done in timerEvent() - is it possible with QCanvasXyz directly?
	gnat->setVelocity( xmov * gnatSpeed,ymov * gnatSpeed);
	if ( rnd == 3 ) {
		gnat->move( 0, height() );
	} else {
		switch ( random()%4 ) {
			case 0:
				gnat->move( 0 - gnat->boundingRect().width(), 0  - gnat->boundingRect().height() );
				break;
			case 1:
				gnat->move( width(), 0 - gnat->boundingRect().height() );
				break;
			case 2:
				gnat->move( width(), height() );
				break;
			default:
				gnat->move( 0 - gnat->boundingRect().width(), this->height() );
				break;
		}
	}
 gnats.append( gnat );
 }
}


void GameField::addMag( int num )
{
 int i;

 for ( i = 0; i < num; i++ ) {
	Mag *mag = new Mag( magImg, field );
	
//	mag->setBoundsAction( QwRealMobileSprite::Wrap );
	mag->setVelocity( 1 + 8/*from forward*/, 0 );
	mag->move( 0,this->height() - 20 );
	mag->setVisible(true);

	mags.append( mag );
 }
}

void GameField::addTux( int num )
{
 for ( int i = 0; i < num; i++ ) {
	Tux *tux = new Tux( tuxImg, field );
	
//	tux->setBoundsAction( QwRealMobileSprite::Wrap );
	tux->setVelocity( 1 + 9, 0 );
	tux->move( 0,this->height()-90, 0 );
	tux->setVisible(true);

	tuxs.append( tux );
 }
}

// This will test the position of the shoot, and kill the gnat
void GameField::shoot(QPoint pos)
{
// Check, if a new game is started
 if ( !started )
	return;
	
 if (bulletTimer->isActive())
	return;

 if ( shootWait < 10 ) {
	return;
 } else {
	shootWait = 0;
 }

 if ( shots >= 10 ) {
	if ( sound ) {
		KAudioPlayer::play(locate("appdata", EMPTY_WAV));
	}
	return;
 }
 shots++;

 if ( sound ) {
	KAudioPlayer::play(locate("appdata", SHOT_WAV));
 }


 QCanvasItemList list = field->collisions(pos);
 if (!list.isEmpty()) {
	QCanvasSprite* item = (QCanvasSprite*)list.first();
	if (bullets.contains(item)) {
		if (list.count() > 1)
		item = (QCanvasSprite*)list[1];
	}
	if (gnats.contains(item)) {
		gnats.removeRef( item );
		emit gnatKilled(true);
		bulletTimer->start(40);
		addGnats(1);
		if ( random()%10 == 3 )
			addMag(1);
		if ( random()%35 == 3 )
			addTux(1);
		showDeadGnat(pos.x(),pos.y());
		return;
	} else if (mags.contains(item)) {
		mags.removeRef( item );
		emit magKilled(true);
		bulletTimer->start(40);
		showDeadGnat(pos.x(),pos.y());
		return;
	} else if (tuxs.contains(item)) {
		tuxs.removeRef( item );
		emit tuxKilled(true);
		bulletTimer->start(40);
		showDeadTux(pos.x(),pos.y());
		return;
	}
 }
 emit gnatKilled(false);

 bulletTimer->start(40);

/*
	QListIterator<QCanvasSprite> it(gnats);
	while (it.current()) {
		QCanvasSprite *gnat = it.current();
		if (x >= gnat->x() && x <= gnat->x() + 74 &&
				y >= gnat->y() && y <= gnat->y() + 40) {
			gnats.removeRef( gnat );
			emit gnatKilled(true);
			bulletTimer->start(40);
			addGnats(1);
			if ( random()%10 == 3 )
				addMag(1);
			if ( random()%35 == 3 )
				addTux(1);
			showDeadGnat(x,y);
			return;
		}
		++it;
	}
	
	it = mags;
	while (it.current()) {
		QCanvasSprite *mag = it.current();
		if (x >= mag->x() + 5 && x <= mag->x() + 30 &&
				y >= mag->y() + 5 && y <= mag->y() + 20) {
			mags.removeRef( mag );
			emit magKilled(true);
			bulletTimer->start(40);
			showDeadGnat(x,y);
			return;
		}

		++it;
	}

	it = tuxs;
	while (it.current()) {
		// Killed TUX ??? :-O
		QCanvasSprite *tux = it.current();
		if (x >= tux->x() + 3 && x <= tux->x() + 40 &&
				y >= tux->y() + 5 && y <= tux->y() + 80) {
			tux->hide();
			tuxs.removeRef( tux );
			emit tuxKilled(true);
			bulletTimer->start(40);
			showDeadTux(x,y);
			return;
		}
		++it;
	}
	
	emit gnatKilled(false);
	bulletTimer->start(40);*/
}

void GameField::showDeadGnat(int x, int y)
{
 KDeadGnat *dgnat = new KDeadGnat(GnatKillImg, field);
 dgnat->setVelocity( 0,0 );
 dgnat->move(x - 20 , y - 20);
 dgnat->setVisible(true);
 dgnats.append(dgnat);
}

void GameField::showDeadTux(int x, int y)
{
 DeadTux *dtux = new DeadTux(TuxKillImg, field);

// dtux->setBoundsAction( QwRealMobileSprite::Wrap );//FIXME->unused?
 dtux->setVelocity( 0,0 );
 dtux->move(x-20 , y-20);
 dtux->setVisible(true);

 dtuxs.append(dtux);
}

void GameField::removeDGnat()
{
 QCanvasSprite *dgnat;

 // WHere are the gnats ???
 for ( dgnat = dgnats.first(); dgnat; dgnat = dgnats.next() ) {
	dgnat->hide();
	dgnats.removeRef( dgnat );
 }
 fc = 0;
}

void GameField::removeDTux()
{
 QCanvasSprite *dtux;

 // WHere are the gnats ???
 for ( dtux = dtuxs.first(); dtux; dtux = dtuxs.next() ) {
	dtux->hide();
	dtuxs.removeRef( dtux );
 }
 ft = 0;
}

void GameField::reloadGun()
{
// Check, if a new game is started
 if ( !started )
	return;

 if ( shots < 10 ) {
	return;
 } else {
	if ( sound ) {
		KAudioPlayer::play(locate("appdata", LOAD_WAV));
	}
	shots = 0;
	emit reload();//unused
	addBullets(10);
 }
}

